#pragma once
#include <windows.h>

#define USE_RTTI // Use run time type information in asserts.

#define cexport extern "C" __declspec( dllexport )

#define FImplies( x, y )		( !( x ) || ( y ) )
#define DimensionOf( x )		( sizeof( x ) / sizeof( *( x ) ) )
#define RgC( s )				s, DimensionOf( s )

template <class T> T Min(T a, T b) { return (a < b) ? a : b; }
template <class T> T Max(T a, T b) { return (a > b) ? a : b; }
template <class T> T Abs(T x) { return (x < 0) ? -x : x; }

#include "dbg.h"

#ifdef USE_RTTI
#include "typeinfo.h"
#endif

#define OVERRIDE
#if defined( _MSC_VER )
#if _MSC_VER >= 1400
#undef OVERRIDE
#define OVERRIDE override
#endif
#endif

/*
 * Critical sections.
 */

class CritSec
{
public:
					CritSec() { InitializeCriticalSection( &m_cs ); }
					~CritSec() { DeleteCriticalSection( &m_cs ); }

	void			Enter() { EnterCriticalSection( &m_cs ); }
	void			Leave() { LeaveCriticalSection( &m_cs ); }

private:
	CRITICAL_SECTION m_cs;
};

class ScopedCritSec
{
public:
					ScopedCritSec( CritSec& cs ) : m_pcs( &cs ) { m_pcs->Enter(); }
					~ScopedCritSec() { m_pcs->Leave(); }

	void			Enter() { m_pcs->Enter(); }
	void			Leave() { m_pcs->Leave(); }

private:
	CritSec* const	m_pcs;

	ScopedCritSec& operator=( ScopedCritSec const& ); // Forbidden.
	ScopedCritSec( ScopedCritSec const& ); // Forbidden.
};

/*
 * Refcounted object base class.
 */

class RefCounted
{
public:
					RefCounted() : m_cRef( 1 ) {}
	virtual			~RefCounted() { Assert( !m_cRef ); }

	ULONG			AddRef();
	ULONG			Release();

#ifdef DEBUG
	LONG			GetRefCount() const { return m_cRef; }
#endif

private:
	volatile LONG	m_cRef;
};

/*
 * Smart pointers.
 */

#ifdef USE_RTTI
#define TypeIdName( x )			typeid( x ).name()
#define TypeIdName2( x, y )		typeid( x, y ).name()
#define TypeIdName3( x, y, z )	typeid( x, y, z ).name()
#else
#define TypeIdName( x )			#x
#define TypeIdName2( x, y )		#x "," #y
#define TypeIdName3( x, y, z )	#x "," #y "," #z
#endif

template <class Type>
class SPNew
{
public:
	SPNew( Type* p = 0 )		{ m_p = p; }
	~SPNew()					{ delete m_p; }
	operator Type*() const		{ return m_p; }
	Type& operator*() const		{ Assert( m_p ); return *m_p; }
	Type& operator[]( int i ) const { Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned i ) const { Assert( m_p ); return m_p[i]; }
	Type** operator &()			{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPNew<Type> ) ); return &m_p; }
	Type* operator=( Type* p )	{ Assign( p ); return m_p; }
	void Assign( Type* p )		{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPNew<Type> ) ); return m_p = p; }
	Type* Transfer()			{ Type* p = m_p; m_p = 0; return p; }
	void Free()					{ delete m_p; m_p = 0; }
	Type* Pointer() const		{ return m_p; }
	Type* operator->() const	{ return m_p; }
	bool operator!()			{ return !m_p; }

	Type** UnsafeAddress()		{ return &m_p; }

protected:
	Type* m_p;

private:
	SPNew<Type>& operator=( SPNew<Type> const& ); // Forbidden.
	SPNew( SPNew<Type> const& ); // Forbidden.
};

template <class Type>
class SPMem
{
public:
	SPMem( Type* p = 0 )		{ m_p = p; }
	~SPMem()					{ free( m_p ); }
	operator Type*() const		{ return m_p; }
	Type& operator*() const		{ Assert( m_p ); return *m_p; }
	Type& operator[]( int i ) const { Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned i ) const { Assert( m_p ); return m_p[i]; }
	Type** operator &()			{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPNew<Type> ) ); return &m_p; }
	Type* operator=( Type* p )	{ Assign( p ); return m_p; }
	void Assign( Type* p )		{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPNew<Type> ) ); m_p = p; }
	Type* Transfer()			{ Type* p = m_p; m_p = 0; return p; }
	void Free()					{ free( m_p ); m_p = 0; }
	Type* Pointer() const		{ return m_p; }
	Type* operator->() const	{ return m_p; }
	bool operator!()			{ return !m_p; }

	Type** UnsafeAddress()		{ return &m_p; }

protected:
	Type* m_p;

private:
	SPMem<Type>& operator=( SPMem<Type> const& ); // Forbidden.
	SPMem( SPMem<Type> const& ); // Forbidden.
};

template <class IFace>
class PrivateRelease : public IFace
{
private:
	STDMETHODIMP_( ULONG ) Release(); // Force the Release method to be private to prevent "spfoo->Release()".
};

template <class IFace>
class SPI
{
public:
	SPI()						{ m_p = 0; }
	SPI( IFace* p )				{ m_p = p; }
	~SPI()						{ if ( m_p ) m_p->Release(); }
	operator IFace*() const		{ return m_p; }
	PrivateRelease<IFace>* operator->() const { return static_cast<PrivateRelease<IFace>*>( m_p ); }
	IFace** operator &()		{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPI<IFace> ) ); return &m_p; }
	IFace* operator=( IFace* p ) { Assign( p ); return m_p; }
	void Assign( IFace* p )		{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPI<IFace> ) ); m_p = p; }
	IFace* Transfer()			{ IFace* p = m_p; m_p = 0; return p; }
	IFace* Copy()				{ if ( m_p ) m_p->AddRef(); return m_p; }
	void Release()				{ SetInternal( ( IFace** )&m_p, 0 ); }
	void Set( IFace* p )		{ SetInternal( ( IFace** )&m_p, p ); }
	bool operator!()			{ return ( m_p == 0 ); }

protected:
	void SetInternal( IFace* volatile* ppL, IFace* pR );

protected:
	IFace* m_p;

private:
	SPI<IFace>& operator=( SPI<IFace> const& sp ); // Forbidden.
	SPI( SPI<IFace> const& ); // Forbidden.
};

template <class IFace>
void SPI<IFace>::SetInternal( IFace* volatile* ppL, IFace* pR )
{
	// Addref right side first, in case pR and *ppL are on the same object or
	// are the same variable.

	if( pR )
		pR->AddRef();

	if( *ppL )
	{
		IFace* const pRel = *ppL;
		*ppL = 0;
		pRel->Release();
	}

	*ppL = pR;
}

/*
 * Interface declaration helpers.
 */

#define IMPL

#define DeclareIUnknownMembers( IPURE ) \
	STDMETHOD( QueryInterface ) ( THIS_ REFIID riid, LPVOID* ppvObj ) IPURE; \
	STDMETHOD_( ULONG, AddRef ) ( THIS )  IPURE; \
	STDMETHOD_( ULONG, Release ) ( THIS ) IPURE; \

#define DeclareIRichEditOleCallbackMembers( IPURE ) \
	STDMETHOD( GetNewStorage ) ( THIS_ LPSTORAGE* lplpstg ) IPURE; \
	STDMETHOD( GetInPlaceContext ) ( THIS_ LPOLEINPLACEFRAME* lplpFrame, LPOLEINPLACEUIWINDOW* lplpDoc, LPOLEINPLACEFRAMEINFO lpFrameInfo ) IPURE; \
	STDMETHOD( ShowContainerUI ) ( THIS_ BOOL fShow ) IPURE; \
	STDMETHOD( QueryInsertObject ) ( THIS_ LPCLSID lpclsid, LPSTORAGE lpstg, LONG cp ) IPURE; \
	STDMETHOD( DeleteObject ) ( THIS_ LPOLEOBJECT lpoleobj ) IPURE; \
	STDMETHOD( QueryAcceptData ) ( THIS_ LPDATAOBJECT lpdataobj, CLIPFORMAT* lpcfFormat, DWORD reco, BOOL fReally, HGLOBAL hMetaPict ) IPURE; \
	STDMETHOD( ContextSensitiveHelp ) ( THIS_ BOOL fEnterMode ) IPURE; \
	STDMETHOD( GetClipboardData ) ( THIS_ CHARRANGE* lpchrg, DWORD reco, LPDATAOBJECT* lplpdataobj ) IPURE; \
	STDMETHOD( GetDragDropEffect ) ( THIS_ BOOL fDrag, DWORD grfKeyState, LPDWORD pdwEffect ) IPURE; \
	STDMETHOD( GetContextMenu ) ( THIS_ WORD seltype, LPOLEOBJECT lpoleobj, CHARRANGE* lpchrg, HMENU* lphmenu ) IPURE; \

