/*
 * Utilities.
 */

#include <windows.h>
#include <uxtheme.h>
#include <tmschema.h>
#include "std.h"
#include "util.h"

#pragma warning( disable: 4201 )

/*
 * Resource helpers.
 */

HINSTANCE
ResourceModule()
{
	static HINSTANCE s_hInst = 0;

	if( !s_hInst )
	{
		HMODULE hmodule;
		DWORD const flags = GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS|GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT;
		if( GetModuleHandleEx( flags, LPCSTR( DialogBoxParamU ), &hmodule ) )
			s_hInst = hmodule;
	}

	return s_hInst;
}

/*
 * Theme helpers.
 */

union UxThemeApis
{
	FARPROC Purecall[7];
	struct
	{
		BOOL ( __stdcall* IsThemeActive )( void );
		HTHEME ( __stdcall* OpenThemeData )( HWND hwnd, LPCWSTR pszClassList );
		BOOL ( __stdcall* IsThemeBackgroundPartiallyTransparent )( HTHEME htheme, int iPartId, int iStateId );
		HRESULT ( __stdcall* DrawThemeBackground )( HTHEME htheme, HDC hdc, int iPartId, int iStateId, RECT const* pRect, RECT const* pClipRect );
		HRESULT ( __stdcall* DrawThemeParentBackground )( HWND hwnd, HDC hdc, RECT* prc );
		HRESULT ( __stdcall* CloseThemeData )( HTHEME htheme );
		HRESULT ( __stdcall* SetWindowTheme )( HWND hwnd, LPCWSTR pszSubAppName, LPCWSTR pszSubIdList );
	};
};

static UxThemeApis s_uxtheme = { 0 };
static HINSTANCE s_hUxTheme = 0;
static bool s_fUxThemeLoaded = false;

static char const* const c_rgNames[] =
{
	"IsThemeActive",
	"OpenThemeData",
	"IsThemeBackgroundPartiallyTransparent",
	"DrawThemeBackground",
	"DrawThemeParentBackground",
	"CloseThemeData",
	"SetWindowTheme",
};
CASSERT( DimensionOf( c_rgNames ) == DimensionOf( s_uxtheme.Purecall ) );

static bool
ThemeLoadLibrary()
{
	if( !s_fUxThemeLoaded )
	{
		s_hUxTheme = LoadLibrary( "uxtheme.dll" );
		if( s_hUxTheme )
		{
			for( unsigned ii = DimensionOf( s_uxtheme.Purecall ); ii--; )
			{
				s_uxtheme.Purecall[ii] = GetProcAddress( s_hUxTheme, c_rgNames[ii] );
				if( !s_uxtheme.Purecall[ii] )
				{
					FreeLibrary( s_hUxTheme );
					s_hUxTheme = 0;
					break;
				}
			}
		}
		s_fUxThemeLoaded = true;
	}
	return !!s_hUxTheme;
}

static LRESULT
ThemeWndProc_Edit( HWND const hwnd, UINT const message, WPARAM const wParam, LPARAM const lParam )
{
	Assert( s_hUxTheme );

	if( message == WM_NCPAINT &&
		( GetWindowLong( hwnd, GWL_EXSTYLE ) & WS_EX_CLIENTEDGE ) &&
		s_uxtheme.IsThemeActive() )
	{
		RECT rcThis;
		GetClientRect( hwnd, &rcThis );
		MapWindowPoints( hwnd, 0, LPPOINT( &rcThis ), 2 );

		RECT rcNC;
		GetWindowRect( hwnd, &rcNC );

		OffsetRect( &rcThis, -rcNC.left, -rcNC.top );
		OffsetRect( &rcNC, -rcNC.left, -rcNC.top );

		SCROLLBARINFO sbi = { sizeof( sbi ), 0 };
		::GetScrollBarInfo( hwnd, OBJID_HSCROLL, &sbi );
		bool const fHorz = sbi.rgstate[0] != STATE_SYSTEM_INVISIBLE;
		::GetScrollBarInfo( hwnd, OBJID_VSCROLL, &sbi );
		bool const fVert = sbi.rgstate[0] != STATE_SYSTEM_INVISIBLE;

		//$ review: (chrisant) Are scrollbars always on the right and bottom,
		// or can they sometimes be on the left and/or top?

		if( fVert )
			rcThis.right += GetSystemMetrics( SM_CXVSCROLL );
		if( fHorz )
			rcThis.bottom += GetSystemMetrics(SM_CYHSCROLL);

		// Set up the clipping rect and then call the base richedit to paint
		// its NC area.  The base richedit seems to draw outside our clipping
		// rect if WS_EX_CLIENTEDGE is set.  So we clear it temporarily, but
		// that makes the base richedit incorrectly draw the box where the
		// horizontal and vertical scrollbars meet.  So first we tell the base
		// richedit to draw, and then we draw the real themed border
		// afterwards.  This seems to constrain any flicker to just the box
		// where the horizontal and vertical scrollbars meet, and in practice
		// there is no flicker at all, thanks to Gdi batching.

		HDC const hdc = GetWindowDC( hwnd );
		SaveDC( hdc );

		SetWindowLong( hwnd, GWL_EXSTYLE, GetWindowLong( hwnd, GWL_EXSTYLE ) & ~WS_EX_CLIENTEDGE );

		IntersectClipRect( hdc, rcThis.left, rcThis.top, rcThis.right, rcThis.bottom );
		if( fHorz && fVert )
			ExcludeClipRect( hdc, rcThis.right - GetSystemMetrics( SM_CXVSCROLL ), rcThis.bottom - GetSystemMetrics( SM_CYHSCROLL ), rcThis.right, rcThis.bottom );

		WNDPROC const wndproc = WNDPROC( GetProp( hwnd, "ThemeBaseWndProc" ) );
		CallWindowProc( wndproc, hwnd, message, wParam, lParam );

		SetWindowLong( hwnd, GWL_EXSTYLE, GetWindowLong( hwnd, GWL_EXSTYLE ) | WS_EX_CLIENTEDGE );

		RestoreDC( hdc, -1 );

		// Now draw the themed border.  If both the horizontal and vertical
		// scrollbars are present then do some fancy clipping where they meet,
		// per the issue described above.

		SaveDC( hdc );

		if( fHorz && fVert )
		{
			ExcludeClipRect( hdc, rcThis.left, rcThis.top, rcThis.right, rcThis.bottom - GetSystemMetrics( SM_CYHSCROLL ) );
			ExcludeClipRect( hdc, rcThis.left, rcThis.top, rcThis.right - GetSystemMetrics( SM_CXVSCROLL ), rcThis.bottom );
		}
		else
		{
			ExcludeClipRect( hdc, rcThis.left, rcThis.top, rcThis.right, rcThis.bottom );
		}

		HTHEME const hTheme = s_uxtheme.OpenThemeData( hwnd, L"Edit" );

		if( hTheme )
		{
			if( s_uxtheme.IsThemeBackgroundPartiallyTransparent( hTheme, EP_EDITTEXT, ETS_NORMAL ) )
				s_uxtheme.DrawThemeParentBackground( hwnd, hdc, &rcNC );

			int const state = ( !IsWindowEnabled( hwnd ) ? ETS_DISABLED :
								( GetWindowLong( hwnd, GWL_STYLE ) & ES_READONLY ) ? ETS_READONLY :
								ETS_NORMAL );

			s_uxtheme.DrawThemeBackground( hTheme, hdc, EP_EDITTEXT, state, &rcNC, 0 );
			s_uxtheme.CloseThemeData( hTheme );
		}

		RestoreDC( hdc, -1 );
		ReleaseDC( hwnd, hdc );

		return 0;
	}

	WNDPROC const wndproc = WNDPROC( GetProp( hwnd, "ThemeBaseWndProc" ) );
	return CallWindowProc( wndproc, hwnd, message, wParam, lParam );
}

bool
ThemeSubclassEdit( HWND hwnd )
{
	if( !ThemeLoadLibrary() )
		return false;
	Assert( !GetProp( hwnd, "ThemeBaseWndProc" ) );
	WNDPROC const wndproc = WNDPROC( GetWindowLongPtr( hwnd, GWL_WNDPROC ) );
	SetProp( hwnd, "ThemeBaseWndProc", wndproc );
	SetWindowLongPtr( hwnd, GWL_WNDPROC, LONG_PTR( ThemeWndProc_Edit ) );
	return true;
}

bool
ThemeExplorerTheme( HWND hwnd )
{
	if( !ThemeLoadLibrary() )
		return false;
	if( FAILED( s_uxtheme.SetWindowTheme( hwnd, L"explorer", 0 ) ) )
		return false;
	return true;
}

/*
 * Layout helpers.
 */

inline bool
DeferPosXact::DeferWindowPos( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT ulFlags )
{
	Assert( m_hdwp );
	Assert( IsWindow( hwnd ) );

	if( m_hdwp )
		m_hdwp = ::DeferWindowPos( m_hdwp, hwnd, hwndInsertAfter, x, y, cx, cy, ulFlags );

	return !!m_hdwp;
}

void
DeferPosXact::AdjustSize( HWND hwnd, int dx, int dy )
{
	if( hwnd )
	{
		RECT rc;
		HWND const hwndParent = GetParent( hwnd );
		GetWindowRect( hwnd, &rc );
		MapWindowPoints( 0, hwndParent, reinterpret_cast<POINT*>( &rc ), 2 );
		rc.right += dx;
		rc.bottom += dy;
		DeferWindowPos( hwnd, &rc, SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOCOPYBITS );
	}
}

void
DeferPosXact::AdjustPos( HWND hwnd, int dx, int dy )
{
	if( hwnd )
	{
		RECT rc;
		HWND const hwndParent = GetParent( hwnd );
		GetWindowRect( hwnd, &rc );
		MapWindowPoints( 0, hwndParent, reinterpret_cast<POINT*>( &rc ), 2 );
		rc.left += dx;
		rc.top += dy;
		DeferWindowPos( hwnd, &rc, SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOCOPYBITS );
	}
}

/*
 * Color blending.
 */

int
Luminance( COLORREF cr )
{
	// Luminance range is 0..3570 inclusive.

	return ( 3 * WORD( GetRValue( cr ) ) ) + ( 10 * WORD( GetGValue( cr ) ) ) + WORD( GetBValue( cr ) );
}

inline BYTE
BlendValues( BYTE b1, BYTE b2, BYTE alpha )
{
	// Note that if the values are commonly the same, it may improve
	// performance to bail early.

	return BYTE( ( ( WORD( b1 ) * WORD( 255 - alpha ) ) +
				   ( WORD( b2 ) * WORD( alpha ) ) ) / 255 );
}

/**
 * Blends two colors together.
 *
 * @param cr1		First color.
 * @param cr2		Second color.
 * @param alpha		How strongly the second color should be weighted (zero
 *					means the second color is not used; 255 means the first
 *					color is not used).
 *
 * @return COLORREF	The blended color.
 */
COLORREF
BlendColors( COLORREF cr1, COLORREF cr2, BYTE alpha )
{
	// If the alpha level is commonly zero, it's faster to just return the
	// source pixel than to compute it.  If the alpha is not commonly zero
	// then this optimization would actually hurt performance.

	return RGB( BlendValues( GetRValue( cr1 ), GetRValue( cr2 ), alpha ),
				BlendValues( GetGValue( cr1 ), GetGValue( cr2 ), alpha ),
				BlendValues( GetBValue( cr1 ), GetBValue( cr2 ), alpha ) );
}

COLORREF
MakeBlendedGray( HDC hdc )
{
	return MakeBlendedGray( hdc, 0xaa, true );
}

COLORREF
MakeBlendedGray( HDC hdc, BYTE alpha, bool fFavorLight )
{
	COLORREF const crText = GetTextColor( hdc );
	COLORREF const crBk = GetBkColor( hdc );

	// When fFavorLight is true, the caller expresses the alpha in terms of
	// the standard colors (text=black, bk=white) so that the alpha describes
	// how far to blend toward the lighter of the two colors.  When
	// fFavorLight is false, the alpha describes how far to blend toward the
	// background color.

	if( fFavorLight && Luminance( crText ) > Luminance( crBk ) )
		alpha = 255 - alpha;

	return BlendColors( crText, crBk, alpha );
}

